"""
Salt execution module for VMC distributed firewall rules
Provides methods to Create, Read, Update and Delete distributed firewall rules.
"""
import logging
import os

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates

log = logging.getLogger(__name__)

__virtualname__ = "vmc_distributed_firewall_rules"
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def _create_payload_for_distributed_firewall_rule(rule_id, user_input):
    """
    This function creates the payload based on the template and user input passed
    """
    data = vmc_request.create_payload_for_request(
        vmc_templates.create_distributed_firewall_rules, user_input
    )
    data["id"] = data["display_name"] = rule_id
    return data


def list_(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_policy_id,
    verify_ssl=True,
    cert=None,
    cursor=None,
    page_size=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Retrieves distributed firewall rules for Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_distributed_firewall_rules.list hostname=nsxt-manager.local domain_id=cgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the distributed firewall rules should be retrieved

    domain_id
        The domain_id for which the distributed firewall rules should be retrieved. Possible values: mgw, cgw

    security_policy_id
        Security policy id for which rule should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order. Enabled by default.

    """
    log.info("Retrieving %s distributed firewall rules for SDDC %s", domain_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/security-policies/{security_policy_id}/rules"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_policy_id=security_policy_id,
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_distributed_firewall_rules.list",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_id(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_policy_id,
    rule_id,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves given distributed firewall rule from the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_distributed_firewall_rules.get_by_id hostname=nsxt-manager.local domain_id=cgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the distributed firewall rule should be retrieved

    domain_id
        The domain_id for which the distributed firewall rule should be retrieved. Possible values: mgw, cgw

    security_policy_id
        Security policy id for which rule should be retrieved

    rule_id
        The distribute firewall rule id, any static unique string identifying the rule.
        Also same as the display_name by default.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    """
    log.info("Retrieving distributed firewall rule %s for SDDC %s", rule_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/security-policies/{security_policy_id}/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_policy_id=security_policy_id,
        rule_id=rule_id,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_distributed_firewall_rules.get_by_id",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_policy_id,
    rule_id,
    verify_ssl=True,
    cert=None,
):
    """
    Deletes given distributed firewall rule from the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_distributed_firewall_rules.delete hostname=nsxt-manager.local domain_id=cgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the distributed firewall rule should be deleted

    domain_id
        The domain_id for which the distributed firewall rule should be deleted. Possible values: mgw, cgw

    security_policy_id
        Security policy id for which rule should be deleted

    rule_id
        The distribute firewall rule id, any static unique string identifying the rule.
        Also same as the display_name by default.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    """
    log.info("Deleting distributed firewall rule %s for SDDC %s", rule_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/security-policies/{security_policy_id}/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_policy_id=security_policy_id,
        rule_id=rule_id,
    )

    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_distributed_firewall_rules.delete",
        responsebody_applicable=False,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def create(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_policy_id,
    rule_id,
    verify_ssl=True,
    cert=None,
    source_groups=None,
    destination_groups=None,
    services=None,
    scope=None,
    action=None,
    sequence_number=None,
    disabled=None,
    logged=None,
    description=None,
    direction=None,
    notes=None,
    tag=None,
    tags=vmc_constants.VMC_NONE,
):
    """
    Creates Distributed firewall rule for the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_distributed_firewall_rules.create hostname=nsxt-manager.local domain_id=cgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the Distributed firewall rule should be added

    domain_id
        The domain_id for which the Distributed firewall rule should be added. Possible values: default, cgw

    security_policy_id
        Security policy id for which rule should be added

    rule_id
        Id of the distribute firewall rule to be added to given SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    source_groups
        (Optional) List of Source group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    destination_groups
        (Optional) List of Destination group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    services
        (Optional) Names of services. In order to specify all services, use the constant "ANY".
        This is case insensitive. If "ANY" is used, it should be the ONLY element in the services array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    scope
        (Optional) The list of policy paths where the rule is applied LR/Edge/T0/T1/LRP etc.
        Note that a given rule can be applied on multiple LRs/LRPs.

    action
        (Optional) The action to be applied to all the services.
        Possible Values for are: ALLOW, DROP, REJECT

    sequence_number
        (Optional) Sequence number of the Rule.
        This field is used to resolve conflicts between multiple Rules under Security or Gateway Policy for a Domain.
        If no sequence number is specified by the user, a value of 0 is assigned by default.
        If there are multiple rules with the same sequence number then their order is not deterministic.
        If a specific order of rules is desired, then one has to specify unique sequence numbers.

    Disabled
        (Optional) Flag to disable the rule. Default is false.

    logged
        (Optional) Enable logging flag. Flag to enable packet logging. Default is disabled.

    description
        (Optional) Description of this resource

    direction
        (Optional) Define direction of traffic
        Possible Values for are: IN, OUT, IN_OUT
        Default: "IN_OUT"

    notes
        (Optional) Text for additional notes on changes.

    tag
        (Optional) Tag applied on the rule. User level field which will be printed in CLI and packet logs.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code-block::

            {
                "tag": <tag>,
                "scope": <scope>
            }

    Example Values:

       .. code-block::

           {
              "description": " comm entry",
              "sequence_number": 1,
              "source_groups": [
                  "ANY"
              ],
              "logged": false,
              "destination_groups": [
                  "ANY"
              ],
              "scope": [
                  "ANY"
              ],
              "action": "DROP",
              "services": [
                  "ANY"
              ],
              "direction": "IN_OUT",
              "tag": "",
              "notes": ""
            }

    Notes:
        For domain_id cgw
        Rule must have valid values for source, and destination: either source or destination should have valid groups

    Please refer the `Distributed firewall rules <https://developer.vmware.com/apis/nsx-vmc-policy/latest/policy/api/v1/infra/domains/domain-id/security-policies/security-policy-id/rules/rule-id/put/>`_ to get insight of input parameters

    """
    log.info("Creating Distributed firewall rule %s for SDDC %s", rule_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/security-policies/{security_policy_id}/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_policy_id=security_policy_id,
        rule_id=rule_id,
    )

    allowed_dict = {
        "source_groups": source_groups,
        "destination_groups": destination_groups,
        "services": services,
        "scope": scope,
        "action": action,
        "sequence_number": sequence_number,
        "disabled": disabled,
        "logged": logged,
        "description": description,
        "direction": direction,
        "notes": notes,
        "tag": tag,
        "tags": tags,
    }

    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = _create_payload_for_distributed_firewall_rule(rule_id, req_data)
    return vmc_request.call_api(
        method=vmc_constants.PUT_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_distributed_firewall_rules.create",
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def update(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_policy_id,
    rule_id,
    verify_ssl=True,
    cert=None,
    source_groups=None,
    destination_groups=None,
    services=None,
    scope=None,
    action=None,
    sequence_number=None,
    display_name=None,
    disabled=None,
    logged=None,
    description=None,
    direction=None,
    notes=None,
    tag=None,
    tags=vmc_constants.VMC_NONE,
):
    """
    Updates Distributed firewall rule for the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_distributed_firewall_rules.update hostname=nsxt-manager.local domain_id=cgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the Distributed firewall rule belongs to

    domain_id
        The domain_id for which the Distributed firewall rule belongs to. Possible values: default, cgw

    rule_id
        Id of the distributed firewall rule to be updated for given SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    source_groups
        (Optional) List of Source group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    destination_groups
        (Optional) List of Destination group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    services
        (Optional) Names of services. In order to specify all services, use the constant "ANY".
        This is case insensitive. If "ANY" is used, it should be the ONLY element in the services array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    scope
        (Optional) The list of policy paths where the rule is applied LR/Edge/T0/T1/LRP etc.
        Note that a given rule can be applied on multiple LRs/LRPs.

    action
        (Optional) The action to be applied to all the services.
        Possible Values for are: ALLOW, DROP, REJECT

    display_name
        Identifier to use when displaying entity in logs or GUI
            Defaults to ID if not set

    sequence_number
        (Optional) Sequence number of the Rule.
        This field is used to resolve conflicts between multiple Rules under Security or Gateway Policy for a Domain.
        If no sequence number is specified by the user, a value of 0 is assigned by default.
        If there are multiple rules with the same sequence number then their order is not deterministic.
        If a specific order of rules is desired, then one has to specify unique sequence numbers.

    Disabled
        (Optional) Flag to disable the rule. Default is false.

    logged
        (Optional) Enable logging flag. Flag to enable packet logging. Default is disabled.

    description
        (Optional) Description of this resource

    direction
        (Optional) Define direction of traffic
        Possible Values for are: IN, OUT, IN_OUT
        Default: "IN_OUT"

    notes
        (Optional) Text for additional notes on changes.

    tag
        (Optional) Tag applied on the rule. User level field which will be printed in CLI and packet logs.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code-block::

            {
                "tag": <tag>,
                "scope": <scope>
            }

    Example values:

        .. code-block::

           {
              "description": "comm entry",
              "display_name": "",
              "sequence_number": 1,
              "source_groups": [
                "ANY"
              ],
              "logged": false,
              "destination_groups": [
                "ANY"
              ],
              "scope": [
                "ANY"
              ],
              "action": "DROP",
              "services": [
                "ANY"
              ],
              "direction": "IN_OUT",
              "tag": "",
              "notes": ""
            }

    Please refer the `Update Distributed firewall rules <https://developer.vmware.com/apis/nsx-vmc-policy/latest/policy/api/v1/infra/domains/domain-id/security-policies/security-policy-id/rules/rule-id/patch/>`_ to get insight of input parameters

    """
    log.info("Updating Distributed firewall rule %s for SDDC %s", rule_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/security-policies/{security_policy_id}/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_policy_id=security_policy_id,
        rule_id=rule_id,
    )

    # fetch the Distributed firewall rule for the given rule_id
    existing_data = get_by_id(
        hostname,
        refresh_key,
        authorization_host,
        org_id,
        sddc_id,
        domain_id,
        security_policy_id,
        rule_id,
        verify_ssl,
        cert,
    )

    if vmc_constants.ERROR in existing_data:
        return existing_data

    allowed_dict = {
        "source_groups": source_groups,
        "destination_groups": destination_groups,
        "services": services,
        "scope": scope,
        "action": action,
        "display_name": display_name,
        "sequence_number": sequence_number,
        "disabled": disabled,
        "logged": logged,
        "description": description,
        "direction": direction,
        "notes": notes,
        "tag": tag,
        "tags": tags,
    }

    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = vmc_request.create_payload_for_request(
        vmc_templates.update_distributed_firewall_rules, req_data, existing_data
    )
    return vmc_request.call_api(
        method=vmc_constants.PATCH_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_distributed_firewall_rules.update",
        responsebody_applicable=False,
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )



import inspect
from abc import ABC
from abc import abstractmethod

import salt.utils.dictdiffer
from salt.exceptions import SaltInvocationError
from saltext.vmware.utils import nsxt_request


class NSXTPolicyBaseResource(ABC):
    def get(self, url, username, password, **kwargs):

        params = self._create_query_params(**kwargs)

        return nsxt_request.call_api(
            method="get",
            url=url,
            username=username,
            password=password,
            cert_common_name=kwargs.get("cert_common_name"),
            verify_ssl=kwargs.get("verify_ssl", True),
            cert=kwargs.get("cert"),
            params=params,
        )

    def _create_query_params(self, **kwargs):
        query_params = dict()
        for param in self.get_resource_base_query_params():
            if kwargs.get(param):
                query_params[param] = kwargs[param]

        return query_params

    def get_by_display_name(self, url, username, password, display_name, **kwargs):

        results = list()
        page_cursor = None

        while True:
            response_page = self.get(url, username, password, **kwargs, cursor=page_cursor)

            # check if error dictionary is returned
            if "error" in response_page:
                return response_page

            # add all the result from paginated response with given display name to list
            for result in response_page["results"]:
                if result.get("display_name") and result["display_name"] == display_name:
                    results.append(result)

            # if cursor is not present then we are on the last page, end loop
            if "cursor" not in response_page:
                break
            # updated query parameter with cursor
            page_cursor = response_page["cursor"]

        return {"results": results}

    def get_by_display_name_or_id(self, url):

        results = list()
        page_cursor = None

        while True:
            response_page = self.get(
                url,
                username=self.nsx_resource_params.get("username"),
                password=self.nsx_resource_params.get("password"),
                cert=self.nsx_resource_params.get("cert"),
                cert_common_name=self.nsx_resource_params.get("cert_common_name"),
                verify_ssl=self.nsx_resource_params.get("verify_ssl"),
                cursor=page_cursor,
            )

            # check if error dictionary is returned
            if "error" in response_page:
                return response_page

            # filter support based on the display name
            display_name = self.resource_params["display_name"]
            id = self.resource_params["id"]
            if display_name is not None:
                results.extend(
                    result
                    for result in response_page.get("results", [response_page])
                    if result.get("display_name") == display_name
                )

            else:
                results.extend(
                    result
                    for result in response_page.get("results", [response_page])
                    if result.get("id") == id
                )

            # if cursor is not present then we are on the last page, end loop
            if "cursor" not in response_page:
                break
            # updated query parameter with cursor
            page_cursor = response_page["cursor"]

        return {"results": results}

    def get_id_using_display_name(self, url, display_name):
        results = self.get_by_display_name(
            url=url,
            username=self.nsx_resource_params.get("username"),
            password=self.nsx_resource_params.get("password"),
            display_name=display_name,
            cert=self.nsx_resource_params.get("cert"),
            cert_common_name=self.nsx_resource_params.get("cert_common_name"),
            verify_ssl=self.nsx_resource_params.get("verify_ssl"),
        )

        if "error" in results:
            raise SaltInvocationError(
                {"resourceType": self.get_spec_identifier(), "error": results["error"]}
            )

        if not results or len(results.get("results")) == 0:
            raise SaltInvocationError(
                {
                    "resourceType": self.get_spec_identifier(),
                    "error": "No object found with display name {} at path {}".format(
                        display_name, url
                    ),
                }
            )

        if len(results.get("results", [])) > 1:
            raise SaltInvocationError(
                {
                    "resourceType": self.get_spec_identifier(),
                    "error": "Multiple objects found with display name {} at path {}, please provide id".format(
                        display_name, url
                    ),
                }
            )

        return results["results"][0]["id"]

    def create_or_update(
        self,
        hostname,
        username,
        password,
        cert=None,
        cert_common_name=None,
        verify_ssl=True,
        execution_logs=[],
        **kwargs
    ):
        # must call this method for creation of resource and sub-resources

        self.resource_class = self.__class__

        self.nsx_resource_params = self._set_nsx_resource_params(
            hostname=hostname,
            username=username,
            password=password,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )

        self._save(execution_logs, **kwargs)

    @staticmethod
    def get_nsxt_base_url():
        return "https://{}/policy/api/v1"

    def _save(self, execution_logs, **kwargs):
        self.update_resource_params(**kwargs)

        if self.multi_resource_params and len(self.multi_resource_params) > 0:
            for param in self.multi_resource_params:
                self.resource_params = param

                if self.create_or_update_subresource_first(**kwargs):
                    self._patch_subresource(execution_logs, **kwargs)

                is_update = False

                if not hasattr(self, "_parent_info"):
                    self._parent_info = {}
                self.update_parent_info(self._parent_info)

                url = (
                    NSXTPolicyBaseResource.get_nsxt_base_url()
                    + self.get_resource_base_url(self.get_parent_info())
                ).format(self.nsx_resource_params["hostname"])

                get_by_display_name_response = self.get_by_display_name_or_id(url)

                if get_by_display_name_response and "error" in get_by_display_name_response:
                    raise SaltInvocationError(
                        {
                            "resourceType": self.get_spec_identifier(),
                            "error": get_by_display_name_response["error"],
                        }
                    )

                existing_resource = get_by_display_name_response["results"]

                if len(existing_resource) > 1:
                    raise SaltInvocationError(
                        {
                            "resourceType": self.get_spec_identifier(),
                            "error": "More then one resource exist with same name",
                        }
                    )
                else:
                    self.existing_resource_param = (
                        existing_resource[0] if len(existing_resource) > 0 else None
                    )

                if self.existing_resource_param is not None:
                    # If user had specified state as delete for a particular resource it will delete the resource
                    # This flow will not be called for the base resource tier0
                    if (
                        self.resource_params.get("state")
                        and self.resource_params["state"] == "absent"
                    ):
                        self.delete(
                            self.nsx_resource_params["hostname"],
                            self.nsx_resource_params["username"],
                            self.nsx_resource_params["password"],
                            self.existing_resource_param["id"],
                            self.nsx_resource_params.get("cert"),
                            self.nsx_resource_params.get("cert_common_name"),
                            self.nsx_resource_params.get("verify_ssl"),
                            execution_logs,
                        )
                        continue
                    self._fill_missing_resource_params(
                        self.existing_resource_param, self.resource_params
                    )
                    is_update = self._check_for_update(
                        self.existing_resource_param, self.resource_params
                    )
                    if not is_update and not self.create_or_update_subresource_first(**kwargs):
                        self._patch_subresource(execution_logs, **kwargs)
                        continue

                response = self._patch_resource(**kwargs)
                if response:
                    spec_identifier = self.get_spec_identifier()
                    execution_logs.append({"resourceType": spec_identifier, "results": response})

                if not self.create_or_update_subresource_first(**kwargs):
                    self._patch_subresource(execution_logs, **kwargs)

    def _patch_resource(self, **kwargs):

        if not hasattr(self, "_parent_info"):
            self._parent_info = {}
        self.update_parent_info(self._parent_info)

        url = (
            NSXTPolicyBaseResource.get_nsxt_base_url()
            + self.get_resource_base_url(self.get_parent_info())
        ).format(self.nsx_resource_params["hostname"])
        response = self._send_request_to_API(
            resource_base_url=url,
            suffix="/" + self.resource_params["id"],
            method="PATCH",
            data=self.resource_params,
        )

        if response and "error" in response:
            raise SaltInvocationError(
                {"resourceType": self.get_spec_identifier(), "error": response["error"]}
            )

        response = self._send_request_to_API(
            resource_base_url=url, suffix="/" + self.resource_params["id"], method="GET", data=None
        )

        if response and "error" in response:
            raise SaltInvocationError(
                {"resourceType": self.get_spec_identifier(), "error": response["error"]}
            )

        return response

    def update_resource_params(self, resource_params):
        # Can be used to updates the params of resource before making
        # the API call.
        # Should be overridden in the subclass if needed
        pass

    def is_object_deletable(self):
        return True

    def _update_parent_info(self):
        # This update is always performed and should not be overriden by the
        # subresource's class
        self._parent_info["_parent"] = self

    def _patch_subresource(self, execution_logs, **kwargs):
        my_parent = self._parent_info.get("_parent", "")
        self._update_parent_info()
        for sub_resource_class in self._get_sub_resources_class_of(self.resource_class):
            sub_resource = sub_resource_class()
            sub_resource.set_parent_info(self._parent_info)

            sub_resource.create_or_update(
                self.nsx_resource_params["hostname"],
                self.nsx_resource_params["username"],
                self.nsx_resource_params["password"],
                self.nsx_resource_params.get("cert"),
                self.nsx_resource_params.get("cert_common_name"),
                self.nsx_resource_params.get("verify_ssl"),
                execution_logs,
                **kwargs
            )

        self._parent_info["_parent"] = my_parent

    def _get_nsx_access_argument_spec(self):
        return {
            "hostname": {"type": "str", "required": True},
            "username": {"type": "str", "required": True},
            "password": {"type": "str", "required": True},
            "verify_ssl": {"type": "bool", "requried": False, "default": True},
            "cert": {"type": "str", "required": False},
            "cert_common_name": {"type": "str", "required": False},
        }

    def _get_sub_resources_class_of(self, resource_class):
        subresources = []
        for attr in resource_class.__dict__.values():
            if inspect.isclass(attr) and issubclass(attr, NSXTPolicyBaseResource):
                subresources.append(attr)

        # TODO: if update then make reverse as false
        subresources.sort(
            key=lambda subresource: subresource().get_resource_update_priority(), reverse=True
        )
        for subresource in subresources:
            yield subresource

    def _set_nsx_resource_params(self, **resource_params):
        filtered_params = {}

        def filter_with_spec(spec):
            for key in spec.keys():
                if key in resource_params and resource_params[key] is not None:
                    filtered_params[key] = resource_params[key]

        filter_with_spec(self._get_nsx_access_argument_spec())
        return filtered_params

    def update_parent_info(self, parent_info):
        # Override this and fill in self._parent_info if that is to be passed
        # to the sub-resource
        # By default, parent's id is passed
        parent_info[self.get_spec_identifier() + "_id"] = self.resource_params.get("id")

    def _send_request_to_API(
        self,
        suffix="",
        ignore_error=False,
        method="GET",
        data=None,
        resource_base_url=None,
        accepted_error_codes=set(),
    ):

        response = nsxt_request.call_api(
            method,
            resource_base_url + suffix,
            self.nsx_resource_params.get("username"),
            self.nsx_resource_params.get("password"),
            self.nsx_resource_params.get("cert_common_name"),
            self.nsx_resource_params.get("verify_ssl"),
            self.nsx_resource_params.get("cert"),
            data,
            None,
        )
        return response

    @classmethod
    def get_spec_identifier(cls):
        # Can be overriden in the subclass to provide different
        # unique_arg_identifier. It is used to infer which args belong to which
        # subresource.
        # By default, class name is used for subresources.
        return cls.get_resource_name()

    def get_parent_info(self):
        return self._parent_info

    def get_resource_base_query_params(self):
        return ()

    @staticmethod
    @abstractmethod
    def get_resource_base_url(parent_info):
        # Must be overridden by the subclass
        raise NotImplementedError

    @classmethod
    def get_resource_name(cls):
        return cls.__name__

    def create_or_update_subresource_first(self, **kwargs):
        # return True if subresource should be created/updated before parent
        # resource
        return kwargs.get("create_or_update_subresource_first", False)

    @staticmethod
    def get_resource_update_priority():
        # this priority can be used to create/delete subresources
        # at the same level in a particular order.
        # by default, it returns 1 so the resources are created/updated/
        # deleted in a fixed but random order.
        # should be overloaded in subclass to specify its priority.
        # for creation or update, we iterate in descending order.
        # for deletion, we iterate in ascending order.
        return 1

    def set_parent_info(self, parent_info):
        self._parent_info = parent_info

    def delete(
        self,
        hostname,
        username,
        password,
        resource_id,
        cert=None,
        cert_common_name=None,
        verify_ssl=True,
        execution_logs=[],
    ):

        self.resource_class = self.__class__

        self.nsx_resource_params = self._set_nsx_resource_params(
            hostname=hostname,
            username=username,
            password=password,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )

        self._delete_resource_with_id(execution_logs, resource_id)

    def _delete_resource_with_id(self, execution_logs, resource_id):
        if not hasattr(self, "_parent_info"):
            self._parent_info = {}
        self._parent_info[self.get_spec_identifier() + "_id"] = resource_id

        self._delete_sub_resources(execution_logs)

        if self.is_object_deletable():
            url = (
                NSXTPolicyBaseResource.get_nsxt_base_url()
                + self.get_resource_base_url(self.get_parent_info())
            ).format(self.nsx_resource_params["hostname"])
            response = self._send_request_to_API(
                resource_base_url=url, suffix="/" + resource_id, method="DELETE"
            )
            if response and "error" in response:
                raise SaltInvocationError(
                    {"resourceType": self.get_spec_identifier(), "error": response["error"]}
                )
            execution_logs.append(
                {
                    "resourceType": self.get_spec_identifier(),
                    "results": "{} deleted successfully".format(resource_id),
                }
            )

    def _delete_sub_resources(self, execution_logs):
        for sub_resource_class in self._get_sub_resources_class_of(self.resource_class):
            sub_resource = sub_resource_class()
            sub_resource.set_parent_info(self._parent_info)

            url = (
                NSXTPolicyBaseResource.get_nsxt_base_url()
                + sub_resource.get_resource_base_url(self.get_parent_info())
            ).format(self.nsx_resource_params["hostname"])
            response = self._send_request_to_API(resource_base_url=url, suffix="", method="GET")
            if "error" in response:
                raise SaltInvocationError(
                    {"resourceType": sub_resource.get_spec_identifier(), "error": response["error"]}
                )
            if "results" in response:
                for resource in response["results"]:
                    sub_resource.delete(
                        self.nsx_resource_params["hostname"],
                        self.nsx_resource_params["username"],
                        self.nsx_resource_params["password"],
                        resource["id"],
                        self.nsx_resource_params.get("cert"),
                        self.nsx_resource_params.get("cert_common_name"),
                        self.nsx_resource_params.get("verify_ssl"),
                        execution_logs,
                    )
            else:
                sub_resource.delete(
                    self.nsx_resource_params["hostname"],
                    self.nsx_resource_params["username"],
                    self.nsx_resource_params["password"],
                    "bgp",
                    self.nsx_resource_params.get("cert"),
                    self.nsx_resource_params.get("cert_common_name"),
                    self.nsx_resource_params.get("verify_ssl"),
                    execution_logs,
                )

    def get_hierarchy(
        self,
        hostname,
        username,
        password,
        resource_id,
        cert=None,
        cert_common_name=None,
        verify_ssl=True,
        result={},
    ):
        self.resource_class = self.__class__

        self.nsx_resource_params = self._set_nsx_resource_params(
            hostname=hostname,
            username=username,
            password=password,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )
        self._get_hierarchy(resource_id, result)

    def _get_hierarchy(self, resource_id, result):
        if not hasattr(self, "_parent_info"):
            self._parent_info = {}
        self._parent_info[self.get_spec_identifier() + "_id"] = resource_id

        url = (
            NSXTPolicyBaseResource.get_nsxt_base_url()
            + self.get_resource_base_url(self.get_parent_info())
        ).format(self.nsx_resource_params["hostname"])

        response = self._send_request_to_API(
            resource_base_url=url, suffix="/" + resource_id, method="GET", data=None
        )

        if response and "error" in response:
            raise SaltInvocationError(response["error"])

        if self.get_spec_identifier() not in result:
            result[self.get_spec_identifier()] = response
        else:
            children = []
            children.append(result[self.get_spec_identifier()])
            children.append(response)
            result[self.get_spec_identifier()] = children

        self._get_child_hierarchy(response)

    def _get_child_hierarchy(self, result):
        for sub_resource_class in self._get_sub_resources_class_of(self.resource_class):
            sub_resource = sub_resource_class()
            sub_resource.set_parent_info(self._parent_info)

            url = (
                NSXTPolicyBaseResource.get_nsxt_base_url()
                + sub_resource.get_resource_base_url(self.get_parent_info())
            ).format(self.nsx_resource_params["hostname"])
            response = self._send_request_to_API(resource_base_url=url, suffix="", method="GET")
            if response and "error" in response:
                raise SaltInvocationError(
                    "Failure while querying {}: {}".format(
                        sub_resource.get_spec_identifier(), response["error"]
                    )
                )
            if "results" in response:
                for resource in response["results"]:
                    sub_resource.get_hierarchy(
                        self.nsx_resource_params["hostname"],
                        self.nsx_resource_params["username"],
                        self.nsx_resource_params["password"],
                        resource["id"],
                        self.nsx_resource_params.get("cert"),
                        self.nsx_resource_params.get("cert_common_name"),
                        self.nsx_resource_params.get("verify_ssl"),
                        result,
                    )
            else:
                sub_resource.get_hierarchy(
                    self.nsx_resource_params["hostname"],
                    self.nsx_resource_params["username"],
                    self.nsx_resource_params["password"],
                    "bgp",
                    self.nsx_resource_params.get("cert"),
                    self.nsx_resource_params.get("cert_common_name"),
                    self.nsx_resource_params.get("verify_ssl"),
                    result,
                )

    def _check_for_update(self, existing_params, resource_params):
        diff = salt.utils.dictdiffer.deep_diff(existing_params, resource_params)
        return bool(diff)

    def _fill_missing_resource_params(self, existing_params, resource_params):
        """
        resource_params: dict
        existing_params: dict

        Fills resource_params with the key:value from existing_params if
        missing in the former.
        """
        if not existing_params:
            return
        for k, v in existing_params.items():
            if k not in resource_params:
                resource_params[k] = v
            elif isinstance(v, dict):
                self._fill_missing_resource_params(v, resource_params[k])



"""
Salt execution module for VMC Direct Connect
Provides methods to Display Direct Connect Information of an SDDC.
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request

log = logging.getLogger(__name__)

__virtualname__ = "vmc_direct_connect"


def __virtual__():
    return __virtualname__


def get_accounts(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves Direct Connect Account information for given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_direct_connect.get_accounts hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the Direct Connect Account information should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving Direct Connect Account information for SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/accounts"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_direct_connect.get_accounts",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_associated_groups(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves Direct Connect Associated Groups information for given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_direct_connect.get_associated_groups hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the Direct Connect Associated Groups information should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving Direct Connect Associated Groups information for SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/associated-groups/"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_direct_connect.get_associated_groups",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_bgp_info(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves Direct Connect BGP related information for given SDDC, including current Autonomous System Number
    of the VGW attached to the VPC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_direct_connect.get_bgp_info hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the Direct Connect BGP related information should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving Direct Connect BGP related information for SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/direct-connect/bgp"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_direct_connect.get_bgp_info",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_bgp_status(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves Direct Connect BGP status information for given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_direct_connect.get_bgp_status hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the Direct Connect BGP status information should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving Direct Connect BGP status information for SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/realized-state/status?intent_path=/infra/direct-connect/bgp"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_direct_connect.get_bgp_status",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_advertised_routes(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves BGP routes that are advertised by Direct Connect from VMC provider to on-premise datacenter.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_direct_connect.get_advertised_routes hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the BGP routes that are advertised by Direct Connect should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving BGP routes that are advertised by Direct Connect of an SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/direct-connect/routes/advertised"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_direct_connect.get_advertised_routes",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_learned_routes(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieve BGP routes that are learned by Direct Connect from on-premise datacenter.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_direct_connect.get_learned_routes hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the BGP routes that are learned by Direct Connect should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving BGP routes that are learned by Direct Connect of an SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/direct-connect/routes/learned"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_direct_connect.get_learned_routes",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_vifs(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves Direct Connect VIFs (Virtual Interface) available in the given SDDC. The response includes all
    non-connected VIFs (with states "available", "down", "pending" and "confirming") and connected VIFs that are
    available to the given SDDC.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_direct_connect.get_vifs hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the Direct Connect VIFs should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving Direct Connect VIFs available in the SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/direct-connect/vifs"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_direct_connect.get_vifs",
        verify_ssl=verify_ssl,
        cert=cert,
    )



# Copyright 2021 VMware, Inc.
# SPDX-License: Apache-2.0
import logging

import salt.exceptions
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.datacenter as utils_datacenter
import saltext.vmware.utils.vsphere as utils_vmware

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_folder"
__func_alias__ = {"list_": "list"}


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def create(folder_name, dc_name, type, service_instance=None, profile=None):
    """
    Creates a folder on a given datacenter.

    folder_name
        Name of folder.

    dc_name
        Name of datacenter where folder will be created.

    type
        (vm, host, datastore, network) Type of folder to be created.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    dc_ref = utils_datacenter.get_datacenter(service_instance, dc_name)
    folder = utils_common.get_mor_by_property(
        service_instance, vim.Folder, folder_name, "name", dc_ref
    )
    if type == "vm":
        dc_ref.vmFolder.CreateFolder(folder_name)
    elif type == "host":
        dc_ref.hostFolder.CreateFolder(folder_name)
    elif type == "datastore":
        dc_ref.datastoreFolder.CreateFolder(folder_name)
    elif type == "network":
        dc_ref.networkFolder.CreateFolder(folder_name)
    else:
        raise salt.exceptions.CommandExecutionError("invalid type")
    return {"status": "created"}


def destroy(folder_name, dc_name, type, service_instance=None, profile=None):
    """
    Destroy a folder on a given datacenter.

    folder_name
        Name of folder.

    dc_name
        Name of datacenter where folder will be Destroyed.

    type
        (vm, host, datastore, network) Type of folder to be destroyed.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    dc_ref = utils_datacenter.get_datacenter(service_instance, dc_name)
    if type == "vm":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.vmFolder
        )
    elif type == "host":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.hostFolder
        )
    elif type == "datastore":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.datastoreFolder
        )
    elif type == "network":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.networkFolder
        )
    else:
        raise salt.exceptions.CommandExecutionError("invalid type")
    folder.Destroy_Task()
    return {"status": "destroyed"}


def rename(folder_name, new_folder_name, dc_name, type, service_instance=None, profile=None):
    """
    Rename a folder on a given datacenter.

    folder_name
        Name of folder.

    new_folder_name
        Name to rename folder.

    dc_name
        Name of datacenter where folder will be renamed.

    type
        (vm, host, datastore, network) Type of folder to be renamed.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    dc_ref = utils_datacenter.get_datacenter(service_instance, dc_name)
    if type == "vm":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.vmFolder
        )
    elif type == "host":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.hostFolder
        )
    elif type == "datastore":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.datastoreFolder
        )
    elif type == "network":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.networkFolder
        )
    else:
        raise salt.exceptions.CommandExecutionError("invalid type")
    folder.Rename_Task(new_folder_name)
    return {"status": "renamed"}


def move(folder_name, destination_folder_name, dc_name, type, service_instance=None, profile=None):
    """
    Move a folder on a given datacenter.

    folder_name
        Name of folder.

    destination_folder_name
        Destination folder for named folder.

    dc_name
        Name of datacenter where folder will be moved.

    type
        (vm, host, datastore, network) Type of folder to be moved.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    dc_ref = utils_datacenter.get_datacenter(service_instance, dc_name)
    if type == "vm":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.vmFolder
        )
        destination = utils_common.get_mor_by_property(
            service_instance, vim.Folder, destination_folder_name, "name", dc_ref.vmFolder
        )
    elif type == "host":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.hostFolder
        )
        destination = utils_common.get_mor_by_property(
            service_instance, vim.Folder, destination_folder_name, "name", dc_ref.hostFolder
        )
    elif type == "datastore":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.datastoreFolder
        )
        destination = utils_common.get_mor_by_property(
            service_instance, vim.Folder, destination_folder_name, "name", dc_ref.datastoreFolder
        )
    elif type == "network":
        folder = utils_common.get_mor_by_property(
            service_instance, vim.Folder, folder_name, "name", dc_ref.networkFolder
        )
        destination = utils_common.get_mor_by_property(
            service_instance, vim.Folder, destination_folder_name, "name", dc_ref.networkFolder
        )
    else:
        raise salt.exceptions.CommandExecutionError("invalid type")
    task = destination.MoveIntoFolder_Task([folder])
    utils_common.wait_for_task(task, folder.name, "move folder")
    return {"status": "moved"}


def list_(service_instance=None, profile=None):
    """
    .. versionadded:: 23.6.29.0rc1

    Returns a list of folders.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    return utils_vmware.list_folders(service_instance)



"""
VMC Networks state module

Add new network/segment, update existing network/segment and delete existing network/segment from an SDDC.

Example usage :

.. code-block:: yaml

    ensure_network:
      vmc_networks.present:
        - hostname: sample-nsx.vmwarevmc.com
        - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
        - authorization_host: console-stg.cloud.vmware.com
        - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
        - sddc_id: b43da080-2626-f64c-88e8-7f31d9d2c306
        - network_id: web-tier
        - verify_ssl: False
        - cert: /path/to/client/certificate

.. warning::

    It is recommended to pass the VMC authentication details using Pillars rather than specifying as plain text in SLS
    files.


"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_state

log = logging.getLogger(__name__)


def present(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
    subnets=vmc_constants.VMC_NONE,
    admin_state=None,
    description=None,
    domain_name=None,
    tags=vmc_constants.VMC_NONE,
    advanced_config=None,
    l2_extension=vmc_constants.VMC_NONE,
    dhcp_config_path=None,
    display_name=None,
):
    """
    Ensure a given network/segment exists for given SDDC

    name
        Indicates the Network/segment id, any unique string identifying the network/segment.
        Also same as the display_name by default.

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the network/segment should be added

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    subnets
        Subnet configuration required for ROUTED or DISCONNECTED segment/network. It is an array of SegmentSubnet.
        Can contain maximum 1 subnet.
        SegmentSubnet can contain the below fields.

        'gateway_address': (string)
            Gateway IP address in CIDR format for both IPv4 and IPv6.

        'dhcp_ranges': (optional)
            DHCP address ranges are used for dynamic IP allocation. Supports address range and CIDR formats.
            First valid host address from the first value is assigned to DHCP server IP address.
            Existing values cannot be deleted or modified, but additional DHCP ranges can be added.

            It is an array of IPElement (which can be a single IP address, IP address range or a Subnet.
            Its type can be of IPv4 or IPv6).

        'dhcp_config': (optional)
            Additional DHCP configuration for current subnet. It is of type SegmentDhcpConfig which can contain
            the below fields.

            'resource_type':
                can be either SegmentDhcpV4Config or SegmentDhcpV6Config.

            'lease_time':
                DHCP lease time in seconds. When specified, this property overwrites lease time
                configured DHCP server config. Minimum is 60, Maximum is 4294967295 and Default is 86400.

            'server_address':
                IP address of the DHCP server in CIDR format. The server_address is mandatory in case this
                segment has provided a dhcp_config_path and it represents a DHCP server config.
                If the resource_type is a SegmentDhcpV4Config, the address must be an IPv4 address.
                If the resource_type is a SegmentDhcpV6Config, the address must be an IPv6 address.
                This address must not overlap the ip-ranges of the subnet, or the gateway address of
                the subnet, or the DHCP static-binding addresses of this segment.

            'dns_servers':
                IP address of DNS servers for subnet. DNS server IP address must belong to the same address
                family as segment gateway_address property. Maximum items: 2

        For ex:

            .. code::

                subnets:
                  - gateway_address: 100.1.1.1/16
                    dhcp_ranges:
                      - 203.0.113.2/24
                    dhcp_config:
                      resource_type: SegmentDhcpV4Config
                      lease_time: '8000'
                      server_address: 100.1.0.0/16
                      dns_servers:
                        - 203.0.113.0

    admin_state
        (Optional) Represents Desired state of the Segment. Possible values: UP, DOWN
        If this value is not passed, then the vmc-nsx server assigns "UP" as default value.

    description
        (Optional) Description of this resource

    domain_name
        (Optional) DNS domain name.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code::

            tags:
              - tag: <tag-key-1>
                scope: <tag-value-1>
              - tag: <tag-key-2>
                scope: <tag-value-2>

    advanced_config
        (Optional) Advanced configuration for Segment.
        It is a json object which can contain the below fields.

        'connectivity': (String) (optional)
            configuration to manually connect (ON) or disconnect (OFF) a Tier1 segment from corresponding Tier1 gateway.
            Only valid for Tier1 Segments. This property is ignored for L2 VPN extended segments when subnets property
            is not specified.
            Possible values: ON, OFF. If not specified, default will be "ON"

            Note: To create a network/segment of type DISCONNECTED, or to disconnect a ROUTED Segment specify the
            connectivity value as "OFF"

        'address_pool_paths': (array of string) (optional)
            Policy path to IP address pools. Maximum items it can contain is 1.

        For ex:

            .. code::

                advanced_config:
                  address_pool_paths: []
                  connectivity: 'ON'

    l2_extension
        Configuration for extending Segment through L2 VPN. This field is mandatory for EXTENDED segment/network.
        It is a json object which can contain the below fields.

        'l2vpn_paths': (array of string)
            Policy paths corresponding to the associated L2 VPN sessions

        'tunnel_id': (int)
            Tunnel ID. Minimum value is 1 and Maximum value is 4093

        For ex:

            .. code::

                l2_extension:
                  l2vpn_paths:
                    - "/infra/tier-0s/vmc/locale-services/default/l2vpn-services/default/sessions/c1373cd0-b2f0-11eb-80f4-d1a84667de41"
                  tunnel_id: '10'

    dhcp_config_path
        (Optional) Policy path to DHCP configuration. Policy path to DHCP server or relay configuration to
        use for all IPv4 & IPv6 subnets configured on this segment.

    display_name
        Identifier to use when displaying entity in logs or GUI

    Example values:

        .. code::

            display_name: web-tier
            subnets:
              - gateway_address: 40.1.1.1/16
                dhcp_ranges:
                  - 40.1.2.0/24
            admin_state: UP
            description: network segment
            domain_name: net.eng.vmware.com
            tags:
              - tag: tag1
                scope: scope1
            advanced_config:
              address_pool_paths: []
              connectivity: 'ON'
            l2_extension:
              dhcp_config_path: "/infra/dhcp-server-configs/default"

    """
    network_id = name

    input_dict = {
        "subnets": subnets,
        "admin_state": admin_state,
        "description": description,
        "domain_name": domain_name,
        "tags": tags,
        "advanced_config": advanced_config,
        "l2_extension": l2_extension,
        "dhcp_config_path": dhcp_config_path,
        "display_name": display_name,
    }

    input_dict = {k: v for k, v in input_dict.items() if v != vmc_constants.VMC_NONE}

    network = __salt__["vmc_networks.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        network_id=network_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in network:
        if "could not be found" in network["error"]:
            network = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=network["error"], result=False
            )

    if __opts__.get("test"):
        log.info("present is called with test option")
        if network:
            return vmc_state._create_state_response(
                name=name, comment="State present will update network {}".format(network_id)
            )
        else:
            return vmc_state._create_state_response(
                name=name, comment="State present will create network {}".format(network_id)
            )

    if network:
        updatable_keys = input_dict.keys()
        is_update_required = vmc_state._check_for_updates(
            network, input_dict, updatable_keys, ["tags", "subnets", "l2_extension"]
        )

        if is_update_required:
            updated_network = __salt__["vmc_networks.update"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                network_id=network_id,
                verify_ssl=verify_ssl,
                cert=cert,
                subnets=subnets,
                admin_state=admin_state,
                description=description,
                domain_name=domain_name,
                tags=tags,
                advanced_config=advanced_config,
                l2_extension=l2_extension,
                dhcp_config_path=dhcp_config_path,
                display_name=display_name,
            )

            if "error" in updated_network:
                return vmc_state._create_state_response(
                    name=name, comment=updated_network["error"], result=False
                )

            updated_network = __salt__["vmc_networks.get_by_id"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                network_id=network_id,
                verify_ssl=verify_ssl,
                cert=cert,
            )

            if "error" in updated_network:
                return vmc_state._create_state_response(
                    name=name, comment=updated_network["error"], result=False
                )

            return vmc_state._create_state_response(
                name=name,
                comment="Updated network {}".format(network_id),
                old_state=network,
                new_state=updated_network,
                result=True,
            )
        else:
            log.info("All fields are same as existing Network %s", network_id)
            return vmc_state._create_state_response(
                name=name, comment="Network exists already, no action to perform", result=True
            )
    else:
        log.info("No Network found with Id %s", network_id)
        created_network = __salt__["vmc_networks.create"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            network_id=network_id,
            verify_ssl=verify_ssl,
            cert=cert,
            subnets=subnets,
            admin_state=admin_state,
            description=description,
            domain_name=domain_name,
            tags=tags,
            advanced_config=advanced_config,
            l2_extension=l2_extension,
            dhcp_config_path=dhcp_config_path,
        )

        if "error" in created_network:
            return vmc_state._create_state_response(
                name=name, comment=created_network["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Created network {}".format(network_id),
            new_state=created_network,
            result=True,
        )


def absent(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
):
    """
    Ensure a given network/segment does not exist on given SDDC

    name
        Indicates the Network/segment id, any unique string identifying the network/segment.

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the network/segment should be deleted

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.
    """
    network_id = name
    log.info("Checking if Network with Id %s is present", network_id)
    network = __salt__["vmc_networks.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        network_id=network_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in network:
        if "could not be found" in network["error"]:
            network = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=network["error"], result=False
            )

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if network:
            return vmc_state._create_state_response(
                name=name, comment="State absent will delete network with ID {}".format(network_id)
            )
        else:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will do nothing as no network found with ID {}".format(
                    network_id
                ),
            )

    if network:
        log.info("Network found with Id %s", network_id)
        deleted_network = __salt__["vmc_networks.delete"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            network_id=network_id,
            verify_ssl=verify_ssl,
            cert=cert,
        )

        if "error" in deleted_network:
            return vmc_state._create_state_response(
                name=name, comment=deleted_network["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Deleted network {}".format(network_id),
            old_state=network,
            result=True,
        )
    else:
        log.info("No Network found with Id %s", network_id)
        return vmc_state._create_state_response(
            name=name, comment="No network found with ID {}".format(network_id), result=True
        )



# SPDX-License-Identifier: Apache-2.0
import functools
import json
import logging

import salt.exceptions
import salt.utils.data
import salt.utils.dictdiffer
import saltext.vmware.modules.esxi as vmware_esxi
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.drift as drift
import saltext.vmware.utils.esxi as utils_esxi

log = logging.getLogger(__name__)

try:
    import pyVmomi
    from pyVmomi import vmodl, vim, VmomiSupport

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_esxi"
__proxyenabled__ = ["vmware_esxi"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def role_present(name, privilege_ids, esxi_host_name=None, service_instance=None, profile=None):
    """
    Ensure role is present on service instance, which may be an ESXi host or vCenter instance.

    role_name
        Role to create/update on service instance. (required).

    privilege_ids
        List of privileges for the role. (required).
        Refer: https://docs.vmware.com/en/VMware-vSphere/7.0/com.vmware.vsphere.security.doc/GUID-ED56F3C4-77D0-49E3-88B6-B99B8B437B62.html
        Example: ['Folder.Create', 'Folder.Delete'].

    esxi_host_name
        Connect to this ESXi host using your pillar's service_instance credentials. (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    """
    log.debug("Running vmware_esxi.role_present")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__,
        profile=profile,
        esxi_host=esxi_host_name,
    )
    role = __salt__["vmware_esxi.get_role"](role_name=name, service_instance=service_instance)
    sys_privs = {"System.Anonymous", "System.Read", "System.View"}
    del_privs = list(set(role.get("privilege_ids", [])) - sys_privs - set(privilege_ids))
    new_privs = list(set(privilege_ids) - set(role.get("privilege_ids", [])) - sys_privs)
    changes = {
        "new": {
            "role_id": role.get("role_id"),
            "role_name": name,
            "privilege_ids": {
                "added": new_privs,
                "removed": del_privs,
                "current": sorted(list(set(privilege_ids) | sys_privs)),
            },
        },
        "old": role,
    }
    if not role:
        if __opts__["test"]:
            ret["comment"] = "Role {} will be created.".format(name)
            ret["result"] = None
        else:
            role = __salt__["vmware_esxi.add_role"](
                role_name=name, privilege_ids=privilege_ids, service_instance=service_instance
            )
            changes["new"]["role_id"] = role["role_id"]
            ret["comment"] = "Role {} created.".format(name)
            ret["result"] = True
            ret["changes"] = changes
    elif not new_privs and not del_privs:
        ret["comment"] = "Role {} is in the correct state".format(name)
        ret["result"] = None
    elif __opts__["test"]:
        ret[
            "comment"
        ] = "Role {} will be updated. {} privileges will be added. {} privileges will be removed.".format(
            name, ",".join(sorted(new_privs)) or "No", ",".join(sorted(del_privs)) or "No"
        )
        ret["result"] = None
    else:
        __salt__["vmware_esxi.update_role"](
            role_name=name, privilege_ids=privilege_ids, service_instance=service_instance
        )
        ret["comment"] = "Role {} updated.".format(name)
        ret["result"] = True
        ret["changes"] = changes
    return ret


def role_absent(name, esxi_host_name=None, service_instance=None, profile=None):
    """
    Ensure role is absent on service instance, which may be an ESXi host or vCenter instance.

    role_name
        Role to delete on service instance. (required).

    esxi_host_name
        Connect to this ESXi host using your pillar's service_instance credentials. (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    """
    log.debug("Running vmware_esxi.role_absent")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__,
        profile=profile,
        esxi_host=esxi_host_name,
    )
    role = __salt__["vmware_esxi.get_role"](role_name=name, service_instance=service_instance)
    if not role:
        ret["comment"] = "Role {} is not present.".format(name)
        ret["result"] = None
    elif __opts__["test"]:
        ret["comment"] = "Role {} will be deleted.".format(name)
        ret["result"] = None
    else:
        __salt__["vmware_esxi.remove_role"](role_name=name, service_instance=service_instance)
        ret["comment"] = "Role {} deleted.".format(name)
        ret["result"] = True
    return ret


def vmkernel_adapter_present(
    name,
    port_group_name,
    dvswitch_name=None,
    vswitch_name=None,
    enable_fault_tolerance=None,
    enable_management_traffic=None,
    enable_provisioning=None,
    enable_replication=None,
    enable_replication_nfc=None,
    enable_vmotion=None,
    enable_vsan=None,
    mtu=1500,
    network_default_gateway=None,
    network_ip_address=None,
    network_subnet_mask=None,
    network_tcp_ip_stack="default",
    network_type="static",
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Ensure VMKernel Adapter exists on matching ESXi hosts.

    name
        The name of the VMKernel interface to update. (required).

    port_group_name
        The name of the port group for the VMKernel interface. (required).

    dvswitch_name
        The name of the vSphere Distributed Switch (vDS) where to update the VMKernel interface.

    vswitch_name
        The name of the vSwitch where to update the VMKernel interface.

    enable_fault_tolerance
        Enable Fault Tolerance traffic on the VMKernel adapter. Valid values: True, False.

    enable_management_traffic
        Enable Management traffic on the VMKernel adapter. Valid values: True, False.

    enable_provisioning
        Enable Provisioning traffic on the VMKernel adapter. Valid values: True, False.

    enable_replication
        Enable vSphere Replication traffic on the VMKernel adapter. Valid values: True, False.

    enable_replication_nfc
        Enable vSphere Replication NFC traffic on the VMKernel adapter. Valid values: True, False.

    enable_vmotion
        Enable vMotion traffic on the VMKernel adapter. Valid values: True, False.

    enable_vsan
        Enable VSAN traffic on the VMKernel adapter. Valid values: True, False.

    mtu
        The MTU for the VMKernel interface.

    network_default_gateway
        Default gateway (Override default gateway for this adapter).

    network_type
        Type of IP assignment. Valid values: "static", "dhcp".

    network_ip_address
        Static IP address. Required if type = 'static'.

    network_subnet_mask
        Static netmask required. Required if type = 'static'.

    network_tcp_ip_stack
        The TCP/IP stack for the VMKernel interface. Valid values: "default", "provisioning", "vmotion", "vxlan".

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Save Adapter:
          vmware_esxi.vmkernel_adapter_present:
            - name: vmk1
            - port_group_name: portgroup1
            - dvsswitch_name: vswitch1
    """
    log.debug("Running vmware_esxi.vmkernel_adapter_present")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )

    adapters = __salt__["vmware_esxi.get_vmkernel_adapters"](
        adapter_name=name,
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
        service_instance=service_instance,
    )
    add_on_hosts = []
    update_on_hosts = []
    for h in hosts:
        if name and adapters[h.name] and name in adapters[h.name]:
            update_on_hosts.append(h.name)
        else:
            add_on_hosts.append(h.name)

    if __opts__["test"]:
        ret[
            "comment"
        ] = "vmkernel adapter {!r} will be created on {} host(s) and updated on {} host(s).".format(
            name, len(add_on_hosts), len(update_on_hosts)
        )
        ret["result"] = None
    elif not add_on_hosts and not update_on_hosts:
        ret[
            "comment"
        ] = "vmkernel adapter {!r} not created/updated on any host. No changes made.".format(name)
        ret["result"] = True
    else:
        hosts_in_error = []
        sample_exception = None
        for action, hosts in [("add", add_on_hosts), ("update", update_on_hosts)]:
            for host in hosts:
                try:
                    func = None
                    if action == "add":
                        func = functools.partial(__salt__["vmware_esxi.create_vmkernel_adapter"])
                    else:
                        func = functools.partial(
                            __salt__["vmware_esxi.update_vmkernel_adapter"], adapter_name=name
                        )
                    ret_save = func(
                        port_group_name=port_group_name,
                        dvswitch_name=dvswitch_name,
                        vswitch_name=vswitch_name,
                        enable_fault_tolerance=enable_fault_tolerance,
                        enable_management_traffic=enable_management_traffic,
                        enable_provisioning=enable_provisioning,
                        enable_replication=enable_replication,
                        enable_replication_nfc=enable_replication_nfc,
                        enable_vmotion=enable_vmotion,
                        enable_vsan=enable_vsan,
                        mtu=mtu,
                        network_default_gateway=network_default_gateway,
                        network_ip_address=network_ip_address,
                        network_subnet_mask=network_subnet_mask,
                        network_tcp_ip_stack=network_tcp_ip_stack,
                        network_type=network_type,
                        datacenter_name=datacenter_name,
                        cluster_name=cluster_name,
                        host_name=host,
                        service_instance=service_instance,
                        profile=profile,
                    )
                    ret["changes"].update(ret_save)
                except salt.exceptions.SaltException as exc:
                    hosts_in_error.append(host)
                    sample_exception = str(exc)
        ret["comment"] = "vmkernel adapter {!r} created on {}, updated on {}.".format(
            name,
            ",".join(sorted(set(add_on_hosts) - set(hosts_in_error))) or "-",
            ",".join(sorted(set(update_on_hosts) - set(hosts_in_error))) or "-",
        )
        ret["result"] = True
        if hosts_in_error:
            ret["comment"] += "erred on {}. Sample exception - {}".format(
                ",".join(sorted(hosts_in_error)), sample_exception
            )
            ret["result"] = False
    return ret


def vmkernel_adapter_absent(
    name,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Ensure VMKernel Adapter exists on matching ESXi hosts.

    name
        The name of the VMKernel interface to update. (required).

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Delete Adapter:
          vmware_esxi.vmkernel_adapter_absent:
            - name: vmk1
    """
    log.debug("Running vmware_esxi.vmkernel_adapter_absent")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )

    adapters = __salt__["vmware_esxi.get_vmkernel_adapters"](
        adapter_name=name,
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
        service_instance=service_instance,
    )
    delete_on_hosts = [h.name for h in hosts if adapters[h.name]]
    if __opts__["test"]:
        ret["comment"] = "vmkernel adapter {!r} will be deleted on {} host(s).".format(
            name, len(delete_on_hosts)
        )
        ret["result"] = None
    elif not delete_on_hosts:
        ret["comment"] = "vmkernel adapter {!r} absent on all hosts. No changes made.".format(name)
    else:
        hosts_in_error = []
        sample_exception = None
        for host in delete_on_hosts:
            try:
                ret_delete = __salt__["vmware_esxi.delete_vmkernel_adapter"](
                    adapter_name=name,
                    datacenter_name=datacenter_name,
                    cluster_name=cluster_name,
                    host_name=host,
                    service_instance=service_instance,
                )
                ret["changes"].update(ret_delete)
            except salt.exceptions.SaltException as exc:
                hosts_in_error.append(host)
                sample_exception = str(exc)
        ret["comment"] = "vmkernel adapter {!r} deleted on {}.".format(
            name, ",".join(sorted(set(delete_on_hosts) - set(hosts_in_error)))
        )
        ret["result"] = True
        if hosts_in_error:
            ret["comment"] += "erred on {}. Sample exception - {}".format(
                ",".join(sorted(hosts_in_error)), sample_exception
            )
            ret["result"] = False
    return ret


def user_present(
    name,
    password,
    description=None,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Add local users_by_host on matching ESXi hosts.

    name
        User to create on matching ESXi hosts. (required).

    password
        Password for the users_by_host. (required).

    description
        Description of the users_by_host. (optional).

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Create User:
          vmware_esxi.user_present:
            - name: local_user
            - password: secret

    """
    log.debug("Running vmware_esxi.user_present")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    create = update = 0
    failed_hosts = []
    diff = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    users_by_host = __salt__["vmware_esxi.get_user"](
        user_name=name,
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
        service_instance=service_instance,
        profile=profile,
    )
    hosts = __salt__["vmware_esxi.list_hosts"](
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
        service_instance=service_instance,
        profile=profile,
    )
    for host in hosts:
        if host in users_by_host:
            # Cannot determine if password is changed. So, when the users_by_host is found on the host, update it.
            diff[host] = {
                "old": {"name": name, "description": users_by_host[host][name]["description"]},
                "new": {"name": name, "description": description},
                "action": "update",
            }
            update += 1
        else:
            diff[host] = {"new": {"name": name, "description": description}, "action": "create"}
            create += 1
    for host in diff.copy():
        if __opts__["test"]:
            ret[
                "comment"
            ] = "User {} will be created on {} host(s) and updated on {} host(s).".format(
                name, create, update
            )
            ret["result"] = None
        elif diff[host]["action"] == "update":
            try:
                __salt__["vmware_esxi.update_user"](
                    user_name=name,
                    password=password,
                    description=description,
                    datacenter_name=datacenter_name,
                    cluster_name=cluster_name,
                    host_name=host,
                    service_instance=service_instance,
                    profile=profile,
                )
            except salt.exceptions.SaltException as exc:
                update -= 1
                failed_hosts.append(host)
                diff.pop(host)
                ret[
                    "comment"
                ] = "User {} created on {} host(s), updated on {} host(s), failed on {} host(s). List of failed host(s) - {}. Sample Error: {}".format(
                    name, create, update, len(failed_hosts), ",".join(sorted(failed_hosts)), exc
                )
                ret["changes"] = diff
                ret["result"] = False
            if not ret["comment"]:
                ret["comment"] = "User {} created on {} host(s) and updated on {} host(s).".format(
                    name, create, update
                )
                ret["changes"] = diff
                ret["result"] = True
        else:
            try:
                __salt__["vmware_esxi.add_user"](
                    user_name=name,
                    password=password,
                    description=description,
                    datacenter_name=datacenter_name,
                    cluster_name=cluster_name,
                    host_name=host,
                    service_instance=service_instance,
                )
            except salt.exceptions.SaltException as exc:
                create -= 1
                failed_hosts.append(host)
                diff.pop(host)
                ret[
                    "comment"
                ] = "User {} created on {} host(s), updated on {} host(s), failed on {} host(s). List of failed host(s) - {}. Sample Error: {}".format(
                    name, create, update, len(failed_hosts), ",".join(sorted(failed_hosts)), exc
                )
                ret["changes"] = diff
                ret["result"] = False
            if not ret["comment"]:
                ret["comment"] = "User {} created on {} host(s) and updated on {} host(s).".format(
                    name, create, update
                )
                ret["changes"] = diff
                ret["result"] = True
    return ret


def user_absent(
    name,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Remove local users_by_host on matching ESXi hosts.

    name
        User to create on matching ESXi hosts. (required).

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Remove User:
          vmware_esxi.user_absent:
            - name: local_user

    """
    log.debug("Running vmware_esxi.user_absent")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    delete = no_user = 0
    failed_hosts = []
    diff = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    users_by_host = __salt__["vmware_esxi.get_user"](
        user_name=name,
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
        service_instance=service_instance,
    )
    hosts = __salt__["vmware_esxi.list_hosts"](
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
        service_instance=service_instance,
    )
    for host in hosts:
        if host in users_by_host:
            delete += 1
            diff[host] = {name: True}
        else:
            no_user += 1
            diff[host] = {name: False}

    for host in diff.copy():
        if __opts__["test"]:
            if not ret["comment"]:
                ret["comment"] = "User {} will be deleted on {} host(s).".format(name, delete)
                ret["result"] = None
        elif diff[host][name]:
            try:
                __salt__["vmware_esxi.remove_user"](
                    user_name=name,
                    datacenter_name=datacenter_name,
                    cluster_name=cluster_name,
                    host_name=host,
                    service_instance=service_instance,
                )
            except salt.exceptions.SaltException as exc:
                delete -= 1
                failed_hosts.append(host)
                diff.pop(host)
                ret[
                    "comment"
                ] = "User {} removed on {} host(s), failed on {} host(s). List of failed host(s) - {}. Sample Error: {}".format(
                    name, delete, len(failed_hosts), ",".join(sorted(failed_hosts)), exc
                )
                ret["changes"] = diff
                ret["result"] = False
            if not ret["comment"]:
                ret["comment"] = "User {} removed on {} host(s).".format(name, delete)
                ret["changes"] = diff
                ret["result"] = True
    if not ret["comment"]:
        ret["comment"] = "User {} doesn't exist on {} host(s).".format(name, no_user)
        ret["result"] = None
    return ret


def maintenance_mode(
    name,
    enter_maintenance_mode,
    timeout=0,
    evacuate_powered_off_vms=False,
    maintenance_spec=None,
    service_instance=None,
):
    """
    Put host into or out of maintenance mode.

    name
        Host IP or HostSystem/ManagedObjectReference (required).

    enter_maintenance_mode
        If True, put host into maintenance mode.
        If False, put host out of maintenance mode.

    timeout
        If value is greater than 0 then task will timeout if not completed with in window (optional).

    evacuate_powered_off_vms
        Only supported by VirtualCenter (optional).
         If True, for DRS will fail unless all powered-off VMs have been manually registered.
         If False, task will successed with powered-off VMs.
         Only relevant if enter_maintenance_mode must be True.

    maintenance_spec
        HostMaintenanceSpec (optional).
         Only relevant if enter_maintenance_mode must be True.

    service_instance
        Use this vCenter service connection instance instead of creating a new one (optional).

    .. code-block:: bash

        salt '*' vmware_esxi.maintenance_mode '192.0.2.117'
    .. code-block:: yaml

        Maintenance Mode:
          vmware_esxi.maintenance_mode:
            - host: '192.0.2.117'
            - enter_maintenance_mode: true
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # check that host is not all ready in maintenance state.
    host_state = __salt__["vmware_esxi.in_maintenance_mode"](
        host=name, service_instance=service_instance
    )
    if (host_state["maintenanceMode"] == "inMaintenance") == enter_maintenance_mode:
        ret["comment"] = f"Already in {'Maintenance' if enter_maintenance_mode else 'Normal'} mode."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["changes"] = {
            "new": f"Host will enter {'Maintenance' if enter_maintenance_mode else 'Normal'} mode."
        }
        ret["comment"] = "These options are set to change."
        return ret

    if enter_maintenance_mode:
        host_state = __salt__["vmware_esxi.maintenance_mode"](
            host=name,
            timeout=timeout,
            evacuate_powered_off_vms=evacuate_powered_off_vms,
            maintenance_spec=maintenance_spec,
            catch_task_error=True,
            service_instance=service_instance,
        )
    else:
        host_state = __salt__["vmware_esxi.exit_maintenance_mode"](
            host=name, timeout=timeout, catch_task_error=True, service_instance=service_instance
        )

    ret["result"] = (host_state["maintenanceMode"] == "inMaintenance") == enter_maintenance_mode
    if ret["result"]:
        ret["changes"] = {
            "new": f"Host entered {'Maintenance' if enter_maintenance_mode else 'Normal'} mode."
        }
    else:
        ret[
            "comment"
        ] = f"Failed to put host {str(name)} in {'Maintenance' if enter_maintenance_mode else 'Normal'} mode."
    return ret


def lockdown_mode(
    name,
    enter_lockdown_mode,
    datacenter_name=None,
    cluster_name=None,
    get_all_hosts=False,
    service_instance=None,
    profile=None,
):
    """
    Pust a hosts into or out of lockdown.

    name
        IP of single host or list of host_names. If wanting to get a cluster just past an empty list (required).

    enter_lockdown_mode
        If True, put host into lockdown mode.
        If False, put host out of lockdown mode (required)

    datacenter_name
        The datacenter name. Default is None (optional).

    host_names
        The host_names to be retrieved. Default is None (optional).

    cluster_name
        The cluster name - used to restrict the hosts retrieved. Only used if
        the datacenter is set.  This argument is optional (optional).

    get_all_hosts
        Specifies whether to retrieve all hosts in the container.
        Default value is False (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_esxi.lockdown_mode '192.0.2.117'
    .. code-block:: yaml

        Lockdown Mode:
          vmware_esxi.lockdown_mode:
            - host: '192.0.2.117'
            - enter_lockdown_mode: true
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    if not isinstance(name, str):
        host_refs = utils_esxi.get_hosts(
            service_instance=service_instance,
            datacenter_name=datacenter_name,
            host_names=name,
            cluster_name=cluster_name,
            get_all_hosts=get_all_hosts,
        )
    else:
        if isinstance(name, vim.HostSystem):
            host_refs = (name,)
        else:
            service_instance = service_instance or connect.get_service_instance(
                config=__opts__, profile=profile
            )
            host_refs = (utils_esxi.get_host(name, service_instance),)

    for ref in host_refs:
        # check that host is not all ready in lock state.
        host_state = __salt__["vmware_esxi.in_lockdown_mode"](
            host=ref, service_instance=service_instance, profile=profile
        )
        if (host_state["lockdownMode"] == "inLockdown") == enter_lockdown_mode:
            ret[
                "comment"
            ] += f"{ref.name} already in {'Lockdown' if enter_lockdown_mode else 'Normal'} mode.\n"
            continue

        if __opts__["test"]:
            ret["result"] = None
            ret["changes"].setdefault("new", []).append(
                f"{ref.name} will enter {'Lockdown' if enter_lockdown_mode else 'Normal'} mode."
            )
            continue

        if enter_lockdown_mode:
            host_state = __salt__["vmware_esxi.lockdown_mode"](
                host=ref, catch_task_error=True, service_instance=service_instance, profile=profile
            )
        else:
            host_state = __salt__["vmware_esxi.exit_lockdown_mode"](
                host=ref, catch_task_error=True, service_instance=service_instance, profile=profile
            )
        ref_results = (host_state["lockdownMode"] == "inLockdown") == enter_lockdown_mode
        if ret["result"]:
            ret["result"] = ref_results
        if ref_results:
            ret["changes"].setdefault("new", []).append(
                f"{ref.name} entered {'Lockdown' if enter_lockdown_mode else 'Normal'} mode."
            )
        else:
            ret[
                "comment"
            ] += f"Failed to put host {ref.name} in {'Lockdown' if enter_lockdown_mode else 'Normal'} mode.\n"
    if ret["result"]:
        ret["comment"] += f"Task was successfully!\n"
    elif ret["result"] is None:
        ret["comment"] += "These options are set to change."
    return ret


def advanced_configs(
    name,
    configs,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Set advanced configuration on matching ESXi hosts.

    configs
        Set of key value pairs to be set on matching ESXi hosts (required)

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)


    .. code-block:: yaml

        ESXi_advanced_config_example:
          vmware_esxi.advanced_configs:
            - configs:
                DCUI.Access: root
                Net.BlockGuestBPDU: 1
    """
    log.debug("Running vmware_esxi.advanced_configs")
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    esxi_config_old = __salt__["vmware_esxi.get_advanced_config"](
        config_name="",
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
        service_instance=service_instance,
    )

    changes = {}

    for key in configs:
        for host in esxi_config_old:
            if key not in esxi_config_old[host]:
                return {
                    "name": name,
                    "result": False,
                    "comment": f"The config {key} does not exist on host {host}",
                    "changes": changes,
                }

    result = True
    comment = "Config already in correct state"
    for host in esxi_config_old:
        diff = salt.utils.data.recursive_diff(
            esxi_config_old[host], configs, ignore_missing_keys=True
        )
        if "new" in diff:
            changes[host] = diff
            if __opts__["test"]:
                if diff:
                    result = None
                    comment = "Changes would be made"
            else:
                config = __salt__["vmware_esxi.set_advanced_configs"](
                    config_dict=diff["new"],
                    datacenter_name=datacenter_name,
                    cluster_name=cluster_name,
                    host_name=host,
                    service_instance=service_instance,
                )
                comment = "Changes were made"
    return {"name": name, "result": result, "comment": comment, "changes": changes}


def firewall_config(
    name,
    value,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
    less=False,
):
    """
    Set firewall configuration on matching ESXi hosts.

    name
        Name of configuration in value. (required).

    value
        Value for configuration on matching ESXi hosts. (required).

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    less
        Default False. If this is set to True, only the changed values will be reported as changes.
    """
    log.debug("Running vmware_esxi.firewall_config")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    if isinstance(value[name], list):
        for i in range(len(value[name])):
            value[name][i] = dict(value[name][i])
            if "allowed_hosts" in value[name][i]:
                value[name][i]["allowed_hosts"] = dict(value[name][i]["allowed_hosts"])

    missing_rules = utils_esxi.get_missing_firewall_rules(value[name], hosts)

    if missing_rules:
        messages = [f"{r[0]} ruleset does not exist on ESXi server {r[1]}." for r in missing_rules]
        comment = "\n".join(messages)
        return {"name": name, "result": False, "comment": comment, "changes": {}}

    old_configs = {}
    for host in hosts:
        old_configs[host.name] = {}
        for ruleset in value[name]:
            rule = ruleset["name"]
            fw_config = utils_esxi.get_firewall_config(
                ruleset_name=rule,
                host_name=host.name,
                service_instance=service_instance,
            )
            old_configs[host.name][rule] = fw_config[host.name][rule]

    if __opts__["test"]:
        for host in hosts:
            ret["changes"][host.name] = {}
            for ruleset in value[name]:
                rule = ruleset["name"]
                ret["changes"][host.name][rule] = {}
                for k in ruleset:
                    if k == "name":
                        continue
                    elif k == "allowed_hosts":
                        for j in ruleset[k]:
                            if old_configs[host.name][rule][k][j] == ruleset[k][j]:
                                if not less:
                                    ret["changes"][host.name][rule][
                                        j
                                    ] = f"{j} is already set to {ruleset[k][j]}"
                            else:
                                if not less:
                                    ret["changes"][host.name][rule][
                                        j
                                    ] = f"{j} will be set to {ruleset[k][j]}"
                                else:
                                    ret["changes"][host.name][rule][j] = old_configs[host.name][
                                        rule
                                    ][k][j]
                    else:
                        if old_configs[host.name][rule][k] == ruleset[k]:
                            if not less:
                                ret["changes"][host.name][rule][
                                    k
                                ] = f"{k} is already set to {ruleset[k]}"
                        else:
                            if not less:
                                ret["changes"][host.name][rule][
                                    k
                                ] = f"{k} will be set to {ruleset[k]}"
                            else:
                                ret["changes"][host.name][rule][k] = old_configs[host.name][rule][k]
        ret["comment"] = "These options are set to change."
        return ret

    ret["result"] = True
    ret["changes"] = {"new": {}, "old": {}}
    ret["comment"] = "Configurations are already in correct state."
    for host in hosts:
        ret["changes"]["new"][host.name] = {}
        ret["changes"]["old"][host.name] = {}
        for ruleset in value[name]:
            rule = ruleset["name"]
            change = False
            ret["changes"]["new"][host.name][rule] = {}
            ret["changes"]["old"][host.name][rule] = {}
            for k in ruleset:
                if k == "name":
                    continue
                ret["changes"]["new"][host.name][rule][k] = {}
                ret["changes"]["old"][host.name][rule][k] = {}
                if k == "allowed_hosts":
                    for j in ruleset[k]:
                        if old_configs[host.name][rule][k][j] != ruleset[k][j]:
                            change = True
                            ret["changes"]["new"][host.name][rule][k][j] = ruleset[k][j]
                            ret["changes"]["old"][host.name][rule][k][j] = old_configs[host.name][
                                rule
                            ][k][j]
                else:
                    if old_configs[host.name][rule][k] != ruleset[k]:
                        change = True
                        ret["changes"]["new"][host.name][rule][k] = ruleset[k]
                        ret["changes"]["old"][host.name][rule][k] = old_configs[host.name][rule][k]
            if change:
                __salt__["vmware_esxi.set_firewall_config"](
                    firewall_config=ruleset,
                    host_name=host.name,
                    service_instance=service_instance,
                )
                ret["comment"] = "Configurations have successfully been changed."

    return ret


def ntp_config(
    name,
    service_running,
    ntp_servers=None,
    service_policy=None,
    service_restart=False,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Set ntp configuration on matching ESXi hosts.

    name
        Name of the state.

    service_running
        Ensures the running state of the ntp daemon for the host. Boolean value where
        ``True`` indicates that ntpd should be running and ``False`` indicates that it
        should be stopped.

    ntp_servers
        A list of servers that should be added to the ESXi host's NTP configuration.

    service_policy
        The policy to set for the NTP service.

        .. note::

            When setting the service policy to ``off`` or ``on``, you *must* quote the
            setting. If you don't, the yaml parser will set the string to a boolean,
            which will cause trouble checking for stateful changes and will error when
            trying to set the policy on the ESXi host.

    service_restart
        If set to ``True``, the ntp daemon will be restarted, regardless of its previous
        running state. Default is ``False``.

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Set firewall config:
          vmware_esxi.ntp_config:
            - service_running: True
            - ntp_servers:
              - 192.174.1.100
              - 192.174.1.200
            - service_policy: 'on'
            - service_restart: True
    """
    log.debug("Running vmware_esxi.ntp_config")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    ntpd = "ntpd"

    ntp_config = __salt__["vmware_esxi.get_ntp_config"](service_instance=service_instance)
    ntp_service_state = __salt__["vmware_esxi.list_services"](
        service_name=ntpd, service_instance=service_instance
    )

    for host in hosts:
        if ntp_servers and set(ntp_servers) != set(ntp_config[host.name]["ntp_servers"]):
            ret["changes"][host.name] = {}
            if not __opts__["test"]:
                response = __salt__["vmware_esxi.set_ntp_config"](
                    ntp_servers=ntp_servers, host_name=host.name, service_instance=service_instance
                )
                error = response.get("Error")
                if error:
                    ret["result"] = False
                    ret["comment"] = "Error: {}".format(error)
                    return ret
            # Set changes dictionary for ntp_servers
            ret["changes"][host.name].update(
                {"ntp_servers": {"old": ntp_config[host.name]["ntp_servers"], "new": ntp_servers}}
            )

        # Configure service_running state
        ntp_running = ntp_service_state[host.name][ntpd]["state"] == "running"
        if service_running != ntp_running:
            if host.name not in ret["changes"]:
                ret["changes"][host.name] = {}
            # Only run the command if not using test=True
            if not __opts__["test"]:
                # Start ntdp if service_running=True
                if service_running is True:
                    response = __salt__["vmware_esxi.service_start"](
                        service_name=ntpd, host_name=host.name, service_instance=service_instance
                    )
                # Stop ntpd if service_running=False
                else:
                    response = __salt__["vmware_esxi.service_stop"](
                        service_name=ntpd, host_name=host.name, service_instance=service_instance
                    )
            ret["changes"][host.name].update(
                {"service_running": {"old": ntp_running, "new": service_running}}
            )
        # Configure service_policy
        if service_policy:
            if service_policy != ntp_service_state[host.name][ntpd]["startup_policy"]:
                if host.name not in ret["changes"]:
                    ret["changes"][host.name] = {}
                # Only run the command if not using test=True
                if not __opts__["test"]:
                    response = __salt__["vmware_esxi.service_policy"](
                        service_name=ntpd,
                        startup_policy=service_policy,
                        host_name=host.name,
                        service_instance=service_instance,
                    )
                ret["changes"][host.name].update(
                    {
                        "service_policy": {
                            "old": ntp_service_state[host.name][ntpd]["startup_policy"],
                            "new": service_policy,
                        }
                    }
                )

        # Restart ntp_service if service_restart=True
        if service_restart:
            if host.name not in ret["changes"]:
                ret["changes"][host.name] = {}
            # Only run the command if not using test=True
            if not __opts__["test"]:
                response = __salt__["vmware_esxi.service_restart"](
                    service_name=ntpd, host_name=host.name, service_instance=service_instance
                )
            ret["changes"][host.name].update(
                {"service_restart": {"old": "", "new": "NTP Daemon Restarted."}}
            )

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "NTP state will change."
        return ret

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "NTP is already in the desired state."
    return ret


def firewall_configs(
    name,
    config,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
    drift_level=0,
):
    """
    Get/Set firewall configuration on matching ESXi hosts based on drift report.

    name
        Name of configuration. (required).

    config
        Map with configuration values. (required).

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    drift_level
        Defines the tree level at which drift changes will be represented in output (optional)

    .. code-block:: yaml

        firewall_rules_test1:
            vmware_esxi.firewall_configs:
            - configs:
                - name: sshServer
                enabled: True
                - name: sshClient
                enabled: True

        firewall_rules_test2:
            mware_esxi.firewall_configs:
            - profile: vcenter
            - drift_level: 1
            - config:
                - name: sshServer
                enabled: true
                allowed_host:
                    all_ip: false
                    ip_address:
                    - 192.168.0.253
                    - 192.168.10.1
                    ip_network:
                    - 192.168.0.0/24
                - name: sshClient
                enabled: true
    """

    # Keep this structure
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    # Connect to VMware service
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    # Get Host/s list
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )

    # Clone config input to a Map.
    # Can be used to transform input to internal objects and do validation if needed
    new_configs = {}
    for rule_config in config:
        # Create full representation of the object, default or empty values
        new_config = {
            "enabled": rule_config["enabled"],
            "allowed_host": {
                "all_ip": True,  # by default is True
                "ip_address": [],  # by default is Empty
                "ip_network": [],  # by default is Empty
            },
        }
        # Transform / Validate input vs object, e.g. allowed_host section
        if "allowed_host" in rule_config:
            if "ip_address" in rule_config["allowed_host"]:
                ip_addresses = rule_config["allowed_host"]["ip_address"]
                if ip_addresses:
                    new_config["allowed_host"]["all_ip"] = False
                    new_config["allowed_host"]["ip_address"] = ip_addresses
            if "ip_network" in rule_config["allowed_host"]:
                ip_networks = rule_config["allowed_host"]["ip_network"]
                if ip_networks:
                    new_config["allowed_host"]["all_ip"] = False
                    new_config["allowed_host"]["ip_network"] = ip_networks
        new_configs[rule_config["name"]] = new_config

    # Get all firewall rules per host,
    # old_configs holds only the rules that are in the scope of interest (provided in argument config_input)
    old_configs = {}
    for host in hosts:
        firewall_config = host.configManager.firewallSystem
        if not firewall_config:
            continue

        ruleset_configs = {}
        for ruleset in firewall_config.firewallInfo.ruleset:
            # filter only interesting rules (provided as arguments config_input)
            if ruleset.key in new_configs.keys():
                # all fields are present in vmomi object, hence also in our object
                ruleset_configs[ruleset.key] = {
                    "enabled": ruleset.enabled,
                    "allowed_host": {
                        "all_ip": ruleset.allowedHosts.allIp,
                        "ip_address": list(ruleset.allowedHosts.ipAddress),
                        "ip_network": [
                            f"{n.network}/{n.prefixLength}"
                            for n in list(ruleset.allowedHosts.ipNetwork)
                        ],
                    },
                }
        old_configs[host.name] = ruleset_configs

    # Find rules changes
    hosts_changes = {}
    for host in hosts:
        rule_diff = drift.drift_report(
            {host.name: old_configs[host.name]}, {host.name: new_configs}, diff_level=0
        )
        rule_diff = json.loads(json.dumps(rule_diff))  # clone object
        if rule_diff is not None and host.name in rule_diff:
            ret["changes"][host.name] = rule_diff[host.name]

            # add changes for process if not dry-run
            if host.name not in hosts_changes:
                hosts_changes[host.name] = []
            new_rules = rule_diff[host.name]["new"]
            for rule_name in new_rules:
                # don't use delta like this - ({"name": rule_name} | new_rules[rule_name]), but:
                hosts_changes[host.name].append({**{"name": rule_name}, **new_configs[rule_name]})

        # it's used only in changes representation and drift report is bigger than first level
        if drift_level > 0:
            rule_diff = drift.drift_report(
                {host.name: old_configs[host.name]},
                {host.name: new_configs},
                diff_level=drift_level,
            )
            if rule_diff is not None and host.name in rule_diff:
                ret["changes"][host.name] = rule_diff[host.name]

    # If it's not dry-run and has changes, then apply changes
    if not __opts__["test"] and hosts_changes:
        comments = {}
        success = True
        for host_name in hosts_changes:
            changes = hosts_changes[host_name]
            for new_rule in changes:
                try:
                    vmware_esxi.set_firewall_config(
                        firewall_config=new_rule,
                        host_name=host_name,
                        service_instance=service_instance,
                        profile=profile,
                    )

                    comments[host_name + " " + new_rule["name"]] = {
                        "status": "SUCCESS",
                        "message": f"Rule '{new_rule['name']}' has been changed successfully for host {host_name}.",
                    }
                except Exception as err:
                    success = False
                    comments[host_name + " " + new_rule["name"]] = {
                        "status": "FAILURE",
                        "message": f"Error occured while setting rule '{new_rule['name']}' for host {host_name}: {err}",
                    }
        # ret["comment"] = "\n ".join(comments)
        ret["comment"] = comments  # it's more readable if passed as object
        ret["result"] = success  # at least one success

    return ret


def password_present(
    name,
    password,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Update the password for a given host.

    name
        Existing user to update on matching ESXi hosts. (required)

    password
        The new password to change on the host. (required)

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Set firewall config:
        vmware_esxi.password_present:
            - name: root
            - password: Password1!
    """
    log.debug("Running vmware_esxi.password_present")
    ret = {
        "name": name,
        "result": True,
        "changes": {},
        "comment": "",
    }

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Host password will change."
        return ret
    else:
        try:
            __salt__["vmware_esxi.update_user"](
                user_name=name,
                password=password,
                datacenter_name=datacenter_name,
                cluster_name=cluster_name,
                host_name=host_name,
                service_instance=service_instance,
                profile=profile,
            )
        except salt.exceptions.CommandExecutionError as err:
            ret["result"] = False
            ret["comment"] = "Error: {}".format(err)
            return ret
    ret["comment"] = "Host password changed."
    return ret


def vsan_config(
    name,
    enabled,
    add_disks_to_vsan=False,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Configures a host's VSAN properties such as enabling or disabling VSAN, or
    adding VSAN-eligible disks to the VSAN system for the host.

    name
        Name of the state.

    enabled
        Ensures whether or not VSAN should be enabled on a host as a boolean
        value where ``True`` indicates that VSAN should be enabled and ``False``
        indicates that VSAN should be disabled.

    add_disks_to_vsan
        If set to ``True``, any VSAN-eligible disks for the given host will be added
        to the host's VSAN system. Default is ``False``.

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    Example:
    .. code-block:: yaml

        configure-host-vsan:
          vmware_esxi.vsan_configured:
            - enabled: True
            - add_disks_to_vsan: True
    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )

    for host in hosts:
        ret["changes"][host.name] = {}
        current_vsan_enabled = __salt__["vmware_esxi.get_vsan_enabled"](
            host_name=host.name, service_instance=service_instance
        )
        error = current_vsan_enabled.get("Error")
        if error:
            ret["comment"] = "Error: {}".format(error)
            return ret
        current_vsan_enabled = current_vsan_enabled.get(host.name)

        # Configure VSAN Enabled state, if changed.
        if enabled != current_vsan_enabled:
            # Only run the command if not using test=True
            if not __opts__["test"]:
                response = __salt__["vmware_esxi.vsan_enable"](
                    enable=enabled, host_name=host.name, service_instance=service_instance
                )
                error = response.get("Error")
                if error:
                    ret["comment"] = "Error: {}".format(error)
                    return ret
            ret["changes"][host.name].update(
                {"enabled": {"old": current_vsan_enabled, "new": enabled}}
            )

        # Add any eligible disks to VSAN, if requested.
        if add_disks_to_vsan:
            current_eligible_disks = __salt__["vmware_esxi.get_vsan_eligible_disks"](
                host_name=host.name, service_instance=service_instance
            )
            error = current_eligible_disks.get("Error")
            if error:
                ret["comment"] = "Error: {}".format(error)
                return ret
            disks = current_eligible_disks.get("Eligible")
            if disks and isinstance(disks, list):
                # Only run the command if not using test=True
                if not __opts__["test"]:
                    response = __salt__["vmware_esxi.vsan_add_disks"](
                        host_name=host.name, service_instance=service_instance
                    )
                    error = response.get("Error")
                    if error:
                        ret["comment"] = "Error: {}".format(error)
                        return ret

                ret["changes"][host.name].update({"add_disks_to_vsan": {"old": "", "new": disks}})

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "VSAN configuration is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "VSAN configuration will change."

    return ret


def vmotion_configured(
    name,
    enabled,
    device="vmk0",
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
):
    """
    Configures a host's VMotion properties such as enabling VMotion and setting
    the device VirtualNic that VMotion will use.

    name
        Name of the state.  (DGM this could have been returned from name of function)

    enabled
        Ensures whether or not VMotion should be enabled on a host as a boolean
        value where ``True`` indicates that VMotion should be enabled and ``False``
        indicates that VMotion should be disabled.

    device
        The device that uniquely identifies the VirtualNic that will be used for
        VMotion for the host. Defaults to ``vmk0``.

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    Example:

    .. code-block:: yaml

        configure-vmotion:
          esxi.vmotion_configured:
            - enabled: True
            - device: sample-device

    """
    log.debug("Running vmware_esxi.firewall_config")
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    try:
        response_list = __salt__["vmware_esxi.get_vmotion_enabled"](
            datacenter_name=datacenter_name,
            cluster_name=cluster_name,
            host_name=host_name,
            service_instance=service_instance,
        )

        # returns a list of host names with enabled or nota
        log.debug(f"DGM vmotion_configured response_list '{response_list}'")
        for host in response_list:
            current_vmotion_enabled = host.get("VMotion_Enabled")

            # DGM work in progress - pausing while helping with 3006 RC1

            # Configure VMotion Enabled state, if changed.
            if enabled != current_vmotion_enabled:
                # Only run the command if not using test=True
                if not __opts__["test"]:
                    # Enable VMotion if enabled=True
                    if enabled is True:
                        ## response = __salt__[esxi_cmd]("vmotion_enable", device=device).get(host)
                        # DGM work to be done for set_vmotion
                        response = __salt__["vmware_esxi.vmotion_enable"](device=device).get(host)
                        error = response.get("Error")
                        if error:
                            ret["comment"] = "Error: {}".format(error)
                            return ret
                    # Disable VMotion if enabled=False
                    else:
                        ## response = __salt__[esxi_cmd]("vmotion_disable").get(host)
                        # DGM work to be done for set_vmotion
                        response = __salt__["vmware_esxi.vmotion_disable"]().get(host)
                        error = response.get("Error")
                        if error:
                            ret["comment"] = "Error: {}".format(error)
                            return ret
                ret["changes"].update(
                    {host: {"enabled": {"old": current_vmotion_enabled, "new": enabled}}}
                )

        ret["result"] = True
        if ret["changes"] == {}:
            ret["comment"] = "VMotion configuration is already in the desired state."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "VMotion configuration will change."

    except salt.exceptions.CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = "Error: {}".format(err)
        return ret

    return ret



"""
Manage VMware NSX-T Data Center Licenses

"""
import json
import logging

from saltext.vmware.utils import nsxt_request

log = logging.getLogger(__name__)

__virtualname__ = "nsxt_license"

NSXT_LICENSE_BASE_URL = "https://{management_host}/api/v1/licenses"


def __virtual__():
    return __virtualname__


def get_licenses(hostname, username, password, verify_ssl=True, cert=None, cert_common_name=None):
    """
    Retrieves license keys from Given NSX-T Manager

    CLI Example:

    .. code-block:: bash

    salt vm_minion nsxt_license.get_licenses hostname=nsxt-manager.local username=admin ...


    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """

    url = NSXT_LICENSE_BASE_URL.format(management_host=hostname)
    log.info("Retrieving license from NSX Manager %s", hostname)

    return nsxt_request.call_api(
        method="get",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def apply_license(
    hostname, username, password, license_key, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Applies given license key to NSX-T Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_license.apply_license hostname=nsxt-manager.local username=admin ...


    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    license_key
        The license key to be added to NSX-T Manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    """
    url = NSXT_LICENSE_BASE_URL.format(management_host=hostname)
    data = {"license_key": license_key}
    log.debug("Applying license to NSX Manager %s", hostname)

    return nsxt_request.call_api(
        method="post",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=data,
    )


def delete_license(
    hostname, username, password, license_key, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Deletes given license key in NSX-T Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_license.delete_license hostname=nsxt-manager.local username=admin ...


    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    license_key
        The license key to be deleted in NSX-T Manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    """
    url = (NSXT_LICENSE_BASE_URL + "?action=delete").format(management_host=hostname)
    data = {"license_key": license_key}
    log.debug("Deleting license from NSX Manager %s", hostname)

    response = nsxt_request.call_api(
        method="post",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=data,
    )

    return response or {"message": "License deleted successfully"}



"""
Salt execution module for Cluster management
Provides methods to Retrieve, Create and Delete cluster in the target SDDC
"""
import logging

from saltext.vmware.modules import vmc_sddc
from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates

log = logging.getLogger(__name__)
__virtualname__ = "vmc_sddc_clusters"
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def list_(hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None):
    """
    Retrieves Clusters list for the given SDDC

    Please refer the `VMC Get SDDC documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

    salt <minion id> vmc_sddc_cluster.get hostname=vmc.vmware.com ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the clusters would be retrieved

    verify_ssl
    (Optional) Option to enable/disable SSL verification. Enabled by default.
    If set to False, the certificate validation is skipped.

    cert
    (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
    The certificate can be retrieved from browser.

    """
    log.info("Retrieving clusters for the sddc %s in the organization %s", sddc_id, org_id)
    sddc_detail = vmc_sddc.get_by_id(
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )
    if "error" in sddc_detail:
        return sddc_detail
    cluster_details = sddc_detail["resource_config"]["clusters"]
    result = {"description": "vmc_sddc_clusters.list_", "clusters": cluster_details}
    return result


def create(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    num_hosts,
    host_cpu_cores_count=None,
    host_instance_type=None,
    msft_license_config=None,
    storage_capacity=None,
    verify_ssl=True,
    cert=None,
):
    """
    Creates a new cluster for a given SDDC with passed config

    Please refer the `VMC create cluster documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/clusters/post/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc_cluster.create hostname=vmc.vmware.com ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the cluster would be created

    num_hosts: Integer
        (Required) Number of hosts in a cluster

    host_cpu_cores_count: Integer
        (Optional) Customize CPU cores on hosts in a cluster.
        Specify number of cores to be enabled on hosts in a cluster

    host_instance_type: enum
        (Optional) The instance type for the esx hosts added to this cluster.
        Possible values are: i3.metal, r5.metal, i3en.metal

    msft_license_config
        (Optional) The desired Microsoft license status to apply to this cluster
        msft_license_config expects to be passed as a dict with provided licences

         .. code-block::

                 msft_license_config": {
                    "mssql_licensing": "string",
                    "windows_licensing": "string"
                }

    storage_capacity: Integer
        (Optional) For EBS-backed instances only, the requested storage capacity in GiB.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    For example:

        .. code-block::

             {
                "host_cpu_cores_count": 0,
                "host_instance_type": "i3.metal",
                "msft_license_config": {
                    "mssql_licensing": "ENABLED",
                    "windows_licensing": "DISABLED",
                    "academic_license": null
                },
                "num_hosts": 1,
                "storage_capacity": 0
            }

    """
    log.info("Creating a new cluster in the SDDC %s", sddc_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs/{sddc_id}/clusters".format(
        base_url=api_base_url, org_id=org_id, sddc_id=sddc_id
    )
    allowed_dict = {
        "num_hosts": num_hosts,
        "host_cpu_cores_count": host_cpu_cores_count,
        "host_instance_type": host_instance_type,
        "msft_license_config": msft_license_config,
        "storage_capacity": storage_capacity,
    }
    req_data = vmc_request._filter_kwargs(allowed_kwargs=allowed_dict.keys(), **allowed_dict)
    request_data = vmc_request.create_payload_for_request(
        vmc_templates.create_sddc_cluster, req_data
    )
    return vmc_request.call_api(
        method=vmc_constants.POST_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc_clusters.create",
        data=request_data,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_primary(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves the primary cluster in provided customer sddc UUID

    Please refer the `VMC Get primary cluster documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/primarycluster/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc_cluster.get_primary_cluster hostname=vmc.vmware.com ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the cluster would be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info(
        "Retrieves the primary cluster for the sddc %s in the organization %s", sddc_id, org_id
    )
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs/{sddc_id}/primarycluster".format(
        base_url=api_base_url, org_id=org_id, sddc_id=sddc_id
    )
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc_clusters.get_primary_cluster",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    cluster_id,
    verify_ssl=True,
    cert=None,
):
    """

    Delete the cluster in the given SDDC
    Please refer the `VMC Delete SDDC cluster documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/clusters/cluster/delete/>`_ to get insight of functionality and input parameters


           This is a force operation which will delete the cluster even if there can be a data loss.
           Before calling this operation, all the VMs should be powered off.

            CLI Example:

            .. code-block:: bash

                salt <minion id> vmc_sddc_cluster.create hostname=vmc.vmware.com ...

            hostname
                The host name of VMC

            refresh_key
                API Token of the user which is used to get the Access Token required for VMC operations

            authorization_host
                Hostname of the Cloud Services Platform (CSP)

            org_id
                The Id of organization to which the SDDC belongs to

            sddc_id
                The Id of SDDC for which the cluster would be deleted

            cluster_id
                The Id of the cluster that would get deleted

            verify_ssl
                (Optional) Option to enable/disable SSL verification. Enabled by default.
                If set to False, the certificate validation is skipped.

            cert
                (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
                The certificate can be retrieved from browser.

    """
    log.info("Deleting a cluster %s in the SDDC %s", cluster_id, sddc_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs/{sddc_id}/clusters/{cluster_id}".format(
        base_url=api_base_url, org_id=org_id, sddc_id=sddc_id, cluster_id=cluster_id
    )
    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc_clusters.delete",
        verify_ssl=verify_ssl,
        cert=cert,
    )



"""
VMC Org User state module

Add new user and delete existing user in the given org.

Example usage :

.. code-block:: yaml

    test_user@vmware.com:
      vmc_org_user.invited:
        - hostname: console-stg.cloud.vmware.com
        - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
        - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
        - organization_roles:
            - name: org_member
        - verify_ssl: False
        - cert: /path/to/client/certificate

.. warning::
    It is recommended to pass the VMC authentication details using Pillars rather than specifying as plain text in SLS
    files.
"""
import logging

from saltext.vmware.utils import vmc_state

log = logging.getLogger(__name__)


def invited(
    name,
    hostname,
    refresh_key,
    org_id,
    organization_roles,
    skip_notify=False,
    custom_roles=None,
    service_roles=None,
    skip_notify_registration=False,
    invited_by=None,
    custom_groups_ids=None,
    verify_ssl=True,
    cert=None,
):
    """
    Ensure a given User exists for given organization.

    name
        Indicates the email identifying the user to be invited to the organization.

    hostname
        The host name of Cloud Services Platform(CSP).

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    org_id
        The ID of organization to which the user belongs to.

    organization_roles
        List of unique organization roles assigned to user.
        It can be specified in the below format

        .. code::

            "organization_roles": [
                        {
                            "name": "org_member"
                        }
                    ]
            where 'name' indicates the name of the organization role.

    skip_notify
        (Optional) Indicates if the notification have to be skipped.

    custom_roles
        (Optional) List of custom role assigned to a user.
        It can be specified in the below format

        .. code::

            "custom_roles": [
                        {
                            "name": "role_name"
                        }
                    ]
            where 'name' indicates the name of the custom role.

    service_roles
        (Optional) List of service roles to attach to a user.
        Below fields defines the properties of service roles.

        'serviceDefinitionLink': (String) (Optional)
            The link to the service definition.

        'serviceRoles': list
            It can be specified in the below format

            .. code::

                "serviceRoles": [
                        {
                            "name": "role_name"
                        }
                    ]
                where 'name' indicates the name of the service role.

    skip_notify_registration
        (Optional) Prevent sending mails to users that do not yet have a CSP profile.

    invited_by
        (Optional) Indicates the actual user who is inviting.

    custom_groups_ids
        (Optional) List of unique IDs of Custom Groups. Can have a maximum of 15 custom group IDs.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    For example:

        .. code-block:: yaml

            test_user@vmware.com:
              vmc_org_user.invite:
                - hostname: console-stg.cloud.vmware.com
                - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
                - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
                - organization_roles:
                    - name: org_member
                - verify_ssl: False
                - cert: /path/to/client/certificate

    """

    username = name
    org_users_list = __salt__["vmc_org_users.list"](
        hostname=hostname,
        refresh_key=refresh_key,
        org_id=org_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in org_users_list:
        return vmc_state._create_state_response(
            name=name,
            comment="Failed to get users for given org : {}".format(org_users_list["error"]),
            result=False,
        )
    org_user = None
    for org_user in org_users_list["results"]:
        if username == org_user["user"].get(
            "username"
        ):  # will the user ever be missing a username?
            break
    else:
        org_user = None

    if __opts__.get("test"):
        log.info("vmc_org_user invite is called with test option")
        if org_user:
            return vmc_state._create_state_response(
                name=name,
                comment="User {} is already part of the organization".format(username),
            )
        else:
            return vmc_state._create_state_response(
                name=name,
                comment="User {} would have been invited".format(username),
            )

    if not org_user:
        invited_org_user = __salt__["vmc_org_users.add"](
            hostname=hostname,
            refresh_key=refresh_key,
            org_id=org_id,
            user_name=username,
            organization_roles=organization_roles,
            skip_notify=skip_notify,
            custom_roles=custom_roles,
            service_roles=service_roles,
            skip_notify_registration=skip_notify_registration,
            invited_by=invited_by,
            custom_groups_ids=custom_groups_ids,
            verify_ssl=verify_ssl,
            cert=cert,
        )

        if "error" in invited_org_user:
            return vmc_state._create_state_response(
                name=name,
                comment="Failed to invite user : {}".format(invited_org_user["error"]),
                result=False,
            )

        return vmc_state._create_state_response(
            name=name,
            comment=invited_org_user["message"],
            new_state=invited_org_user,
            result=True,
        )
    else:
        return vmc_state._create_state_response(
            name=name,
            comment="User {} is already part of the organization".format(username),
            result=True,
        )


def absent(
    name,
    hostname,
    refresh_key,
    org_id,
    notify_users=False,
    verify_ssl=True,
    cert=None,
):
    """
    Ensure a given user does not exist for the given organization.

    name
        Indicates the email identifying the user to be removed from the organization.

    hostname
        The host name of Cloud Services Platform(CSP).

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    org_id
        The ID of organization from which the users should be removed.

    notify_users
        (Optional) Indicates whether the users need to notify through email. Default value is true.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC.
        The certificate can be retrieved from browser.

    For example:

        .. code-block:: yaml

            test_user@vmware.com:
              vmc_org_user.absent:
                - hostname: console-stg.cloud.vmware.com
                - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
                - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
                - verify_ssl: False
                - cert: /path/to/client/certificate

    """

    # search for the users if it's found remove them from the org else say there is no users

    username = name
    log.info("Checking if User with username %s is present in the organization", username)
    org_users_list = __salt__["vmc_org_users.list"](
        hostname=hostname,
        refresh_key=refresh_key,
        org_id=org_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in org_users_list:
        return vmc_state._create_state_response(
            name=name,
            comment="Failed to get users for given org : {}".format(org_users_list["error"]),
            result=False,
        )

    org_user = None
    for org_user in org_users_list["results"]:
        if username == org_user["user"].get("username"):
            user_id = org_user["user"].get("userId")
            log.info("user found with username %s", username)
            break
    else:
        org_user = None

    if org_user:
        if __opts__.get("test"):
            log.info("vmc_org_user.absent is called with test option")
            return vmc_state._create_state_response(
                name=name,
                comment="Would have removed user with username {}".format(username),
            )
        else:
            removed_org_user = __salt__["vmc_org_users.remove"](
                hostname=hostname,
                refresh_key=refresh_key,
                org_id=org_id,
                user_ids=[user_id],
                notify_users=notify_users,
                verify_ssl=verify_ssl,
                cert=cert,
            )

            if "error" in removed_org_user:
                return vmc_state._create_state_response(
                    name=name, comment=removed_org_user["error"], result=False
                )

            return vmc_state._create_state_response(
                name=name,
                comment="Removed user {}".format(username),
                old_state=org_user,
                result=True,
            )
    else:
        log.info(
            "No user found with username %s.",
            username,
        )
        return vmc_state._create_state_response(
            name=name,
            comment="No user found with username {}".format(username),
            result=True,
        )
